MODULE cmor_internal_tables
  
  IMPLICIT NONE
  SAVE
  
  !!  USE cmor_parameters
  
  INTEGER, PARAMETER :: init_undef_integer=-192837
  CHARACTER(len=6), PARAMETER :: init_undef_char='azbycx'
  DOUBLE PRECISION, PARAMETER :: init_undef_double=1.0d20
  REAL, PARAMETER :: init_undef_single=1.0e20
  INTEGER, PARAMETER :: max_requested_coords=200
  INTEGER, PARAMETER :: max_rank=7
  INTEGER, PARAMETER :: a8=8, a16=16, a32=32, a64=64, a128=128, a256=256,  &
                                   a512=512, a1024=1024, a2048=2048
  INTEGER, PARAMETER :: max_num_expts=30
  INTEGER, PARAMETER :: max_mip_tables=20
  INTEGER, PARAMETER :: max_variables=230
  INTEGER, PARAMETER :: max_axes=100
  INTEGER, PARAMETER :: max_dimensions=150
  INTEGER, PARAMETER :: num_axis_args=7  
  INTEGER, PARAMETER :: num_var_args=10
  INTEGER, PARAMETER :: num_write_args=7  
  REAL, PARAMETER :: cmor_version=0.96
  
  ! reserved i/o device numbers
  INTEGER, PARAMETER :: idev_short_log=10
  INTEGER, PARAMETER :: idev_long_log=11
  INTEGER, PARAMETER :: idev_mip_table=12
  
  ! from setup:
  
  CHARACTER(len=a256) :: cmor_inpath
  LOGICAL :: called_setup=.false.
  character(len=a8) :: cmor_netcdf_file_action
  integer :: cmor_set_verbosity
  integer :: cmor_exit_control
  
  
  ! needed:

  INTEGER, TARGET, DIMENSION(num_axis_args) :: axis_arg_cntl
  INTEGER, TARGET, DIMENSION(num_var_args) :: var_arg_cntl
  INTEGER, TARGET, DIMENSION(num_write_args) :: write_arg_cntl
 
  CHARACTER(len=a16), target, DIMENSION(num_axis_args) :: axis_args =  &
       (/ 'table           ', 'table_entry     ', 'units           ', &
          'length          ', 'coord_vals      ', 'cell_bounds     ', &
          'interval        ' /)

  CHARACTER(len=a16), target, DIMENSION(num_var_args) :: var_args =   &
       (/ 'table           ', 'table_entry     ', 'units           ', &
          'axis_ids        ', 'missing_value   ', 'tolerance       ', &
          'positive        ', 'original_name   ', 'history         ', &
          'comment         ' /)

  CHARACTER(len=a16), target, DIMENSION(num_write_args) :: write_args =  &
       (/ 'var_id          ', 'data            ', 'file_suffix     ', &
           'ntimes_passed   ','time_vals       ', 'time_bnds       ', &
           'store_with      '  /)
  
  INTEGER :: num_mip_tables
  INTEGER :: num_axes
  INTEGER :: num_vars
  INTEGER :: num_dims

  !  variable that is currently being treated:
  INTEGER :: active_var_idx
  INTEGER :: active_dim_idx

  !  netCDF id for "label" variable
  INTEGER :: label_id

  TYPE :: mip_table_info
    CHARACTER(len=a8) :: cf_version
    INTEGER :: magic_number
    CHARACTER(len=a128) :: project_id
    CHARACTER(len=a64) :: table_id
    CHARACTER(len=a32) :: table_date
    REAL :: approx_interval
    REAL :: single_missing
    DOUBLE PRECISION :: double_missing
    INTEGER :: integer_missing
    CHARACTER(len=a256+a64+1) :: full_path  
    CHARACTER(len=a512), DIMENSION(max_num_expts) :: expt_id_ok  
  END TYPE mip_table_info

  TYPE :: dataset_info
     
    CHARACTER(len=a512) :: outpath
    CHARACTER(len=a512) :: experiment_id
    CHARACTER(len=a512) :: institution
    CHARACTER(len=a512) :: source
    CHARACTER(len=a32)  :: calendar
    INTEGER :: realization
    CHARACTER(len=a512) :: contact
    CHARACTER(len=a2048) :: history
    CHARACTER(len=a2048) :: comment
    CHARACTER(len=a2048) :: references
    INTEGER :: leap_year
    INTEGER :: leap_month
    INTEGER, DIMENSION(12) :: month_lengths
    INTEGER, DIMENSION(3) :: error_count

  END TYPE dataset_info
  
  TYPE :: variable_table
    CHARACTER(len=a64)   :: mip_table
    INTEGER :: mip_table_idx
    CHARACTER(len=a32)   :: table_entry
    INTEGER, DIMENSION(num_var_args) :: arg_cntl
    CHARACTER(len=a128)  :: standard_name
    CHARACTER(len=a64)   :: out_units
    CHARACTER(len=a256)  :: cell_methods
    CHARACTER(len=a256)  :: long_name
    CHARACTER(len=a2048) :: comment
    CHARACTER(len=a2048) :: history
    CHARACTER(len=a8)    :: out_positive
    INTEGER :: rank
    CHARACTER(LEN=a32), DIMENSION(max_rank) :: axis_names
    INTEGER, DIMENSION(max_rank) :: axis_idxs
    REAL :: valid_min 
    REAL :: valid_max 
    REAL :: ok_min_mean_abs 
    REAL :: ok_max_mean_abs
    CHARACTER(len=a8) :: type
    CHARACTER(len=a32) :: out_name
    INTEGER :: ncid
    INTEGER :: varid
    INTEGER :: nctimid
    INTEGER :: nctim_bndsid
    CHARACTER(len=a512) :: full_outpath
    CHARACTER(len=a16) :: file_status
    CHARACTER(len=a64) :: file_suffix
    INTEGER, DIMENSION(max_rank) :: dim_order
    INTEGER, DIMENSION(max_rank) :: dim_idxs
    DOUBLE PRECISION :: scale_factor
    DOUBLE PRECISION :: offset
    INTEGER :: def_missing
    REAL :: users_missing
    DOUBLE PRECISION :: users_missing_dbl
    INTEGER :: users_missing_int
    REAL :: miss_val_tolerance
    CHARACTER(len=a128) :: users_name
    CHARACTER(len=a64) :: users_units
    CHARACTER(len=a8) :: users_type
    INTEGER, DIMENSION(9) :: error_count
    INTEGER :: times_written
    DOUBLE PRECISION :: last_time_written
    DOUBLE PRECISION :: last_time_bound
    ! The following arrays are needed only if this  variable
    !  is in fact a z factor
    !  zfac_times_written contains the number
    !  of time slices written for a z factor (that is a function of time)
    !  needed by variable i in the variable table
    INTEGER, POINTER, DIMENSION(:) :: zfac_times_written
    REAL, POINTER, DIMENSION(:) :: zfac_values
    DOUBLE PRECISION, POINTER, DIMENSION(:) :: zfac_values_dbl
    REAL, POINTER, DIMENSION(:) :: zfac_bounds
    DOUBLE PRECISION, POINTER, DIMENSION(:) :: zfac_bounds_dbl
  END TYPE variable_table
 
  TYPE axis_table    
    CHARACTER(len=a64) :: mip_table
    CHARACTER(len=a32) :: table_entry
    INTEGER, DIMENSION(num_axis_args) :: arg_cntl
    CHARACTER(len=a128)  :: standard_name
    CHARACTER(len=a64)   :: out_units
    CHaRACTER(LEN=a8) :: axis   ! X, Y, Z, T
    CHARACTER(len=a32)  :: bounds_attrib
    CHARACTER(LEN=a256) :: long_name
    CHARACTER(len=a8)  :: positive_attrib
    INTEGER :: out_direction
    DOUBLE PRECISION :: valid_min
    DOUBLE PRECISION :: valid_max
    CHARACTER(len=a16)  :: type
    CHARACTER(len=a32) :: out_name
    CHARACTER(len=a32) :: coords_attrib
    REAL :: tol_on_requests
    DOUBLE PRECISION :: singleton_val
    CHARACTER(len=a32) :: singleton_label
    DOUBLE PRECISION, DIMENSION(2) :: singleton_bnds
    DOUBLE PRECISION, POINTER, DIMENSION(:) :: requested_vals
    DOUBLE PRECISION, POINTER, DIMENSION(:,:) :: requested_bounds
    CHARACTER(len=a32), POINTER, DIMENSION(:) :: requested_labels
    CHARACTER(len=a256) :: formula
    ChARACTER(len=a32) :: convert_to
    CHARACTER(len=a16), POINTER, DIMENSION(:) :: zfac_names
    CHARACTER(len=a16), POINTER, DIMENSION(:) :: zfac_bnds_names
  END TYPE axis_table
 
  TYPE dimension_table
    INTEGER :: axis_idx
    CHARACTER(len=a64) :: users_units
    CHARACTER(len=a128) :: interval
    LOGICAL :: reverse
    INTEGER :: begin_out
    INTEGER :: lendim
    DOUBLE PRECISION :: scale_factor
    DOUBLE PRECISION :: offset
    ChARACTER(len=a32) :: converted_from
    INTEGER, DIMENSION(6) :: error_count
    INTEGER, POINTER, DIMENSION(:) :: zfac_idxs
    INTEGER, POINTER, DIMENSION(:) :: zfac_bnds_idxs
    DOUBLE PRECISION, POINTER, DIMENSION(:) :: values
    DOUBLE PRECISION, POINTER, DIMENSION(:,:) :: bounds
    CHARACTER(len=a32), POINTER, DIMENSION(:) :: labels   
  END TYPE dimension_table

  TYPE(dataset_info) :: dset
  TYPE(mip_table_info), DIMENSION(max_mip_tables) :: mip_tbl
  TYPE(variable_table), DIMENSION(max_variables) :: var
  TYPE(axis_table), DIMENSION(max_axes) :: axis
  TYPE(dimension_table), DIMENSION(max_dimensions) :: adim

END MODULE cmor_internal_tables

